package tech.riemann.ims.controller.platform.acl;

import club.zhcs.lina.starter.exception.BizException;
import com.baomidou.mybatisplus.core.metadata.IPage;
import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import io.swagger.v3.oas.annotations.Operation;
import io.swagger.v3.oas.annotations.Parameter;
import io.swagger.v3.oas.annotations.tags.Tag;
import lombok.RequiredArgsConstructor;
import org.nutz.lang.Strings;
import org.nutz.lang.random.R;
import org.springframework.http.HttpStatus;
import org.springframework.validation.annotation.Validated;
import org.springframework.web.bind.annotation.*;
import tech.riemann.ims.dto.response.PermissionInfo;
import tech.riemann.ims.entity.acl.Permission;
import tech.riemann.ims.entity.acl.Role;
import tech.riemann.ims.service.acl.IRoleService;

import java.util.List;

/**
 * <p>
 * 角色 前端控制器
 * </p>
 *
 * @author Kerbores(kerbores@gmail.com)
 *
 * @since 2024-07-22 13:56:54
 */
@RestController
@Tag(name = "Role", description = "角色")
@RequiredArgsConstructor
public class RoleController {

    private final IRoleService roleService;

    @GetMapping("roles")
    @Operation(summary = "分页查询角色")
    public IPage<Role> roles(
                             @Parameter(description = "页码") @RequestParam(required = false, defaultValue = "1") long page,
                             @Parameter(description = "页面大小") @RequestParam(required = false, defaultValue = "10") long size,
                             @Parameter(description = "搜索关键词") @RequestParam(required = false, defaultValue = "") String key) {
        boolean hasLike = Strings.isNotBlank(key);
        key = String.format("%%%s%%", key);
        if (hasLike) {
            return roleService.page(Page.<Role> of(page, size),
                                    Wrappers.<Role> lambdaQuery()
                                            .like(hasLike, Role::getName, key)
                                            .or()
                                            .like(Role::getKey, key)
                                            .or()
                                            .like(Role::getDescription, key));
        }
        return roleService.page(Page.<Role> of(page, size), Wrappers.<Role> lambdaQuery());
    }

    @GetMapping("role/{id}")
    @Operation(summary = "角色详情")
    public Role detail(@Parameter(description = "角色id", required = true) @PathVariable long id) {
        return roleService.getById(id);
    }

    @PutMapping("role")
    @Operation(summary = "增加/编辑角色")
    public Role saveOrUpdateRole(@Validated @Parameter(description = "角色") @RequestBody Role role) {
        if (role.getId() != null && role.getId() > 0) {
            if (roleService.update(role, Wrappers.<Role> lambdaUpdate().eq(Role::getId, role.getId()))) {
                return role;
            } else {
                throw BizException.create("更新角色失败!");
            }
        } else {
            role.setKey(R.UU16());
            if (roleService.save(role)) {
                return role;
            } else {
                throw BizException.create("保存角色失败!");
            }
        }
    }

    @DeleteMapping("role/{key}")
    @Operation(summary = "删除角色")
    @ResponseStatus(HttpStatus.OK)
    public void deleteRole(@Parameter(description = "角色key", required = true) @PathVariable String key) {
        if (!roleService.remove(Wrappers.<Role> lambdaQuery().eq(Role::getKey, key))) {
            throw BizException.create("删除角色失败!");
        }
    }

    @GetMapping("role/{key}/permissions")
    @Operation(summary = "查询角色权限")
    public List<Permission> permissions(
                                        @Parameter(description = "角色key") @PathVariable String key) {
        return roleService.permissionsByRoleKey(key);
    }

    @GetMapping("role/{key}/permission-infos")
    @Operation(summary = "查询用于授权的权限信息")
    public List<PermissionInfo> permissionInfos(
                                                @Parameter(description = "角色key") @PathVariable String key) {
        return roleService.permissionInfosByKey(key);
    }

    @PostMapping("role/{key}/permissions")
    @Operation(summary = "为指定角色授权")
    @ResponseStatus(HttpStatus.OK)
    public void grant(
                      @Parameter(description = "角色key") @PathVariable String key,
                      @Parameter(name = "permissions", description = "待授权信息[keyPath]") @RequestBody List<String> permissions) {
        if (!roleService.grant(key, permissions)) {
            throw BizException.create("授权失败");
        }
    }

}
