package tech.riemann.ims.controller.platform.material;

import club.zhcs.lina.starter.exception.BizException;
import com.baomidou.mybatisplus.core.metadata.IPage;
import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import io.swagger.v3.oas.annotations.Operation;
import io.swagger.v3.oas.annotations.Parameter;
import io.swagger.v3.oas.annotations.tags.Tag;
import lombok.RequiredArgsConstructor;
import org.apache.poi.util.StringUtil;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.validation.annotation.Validated;
import org.springframework.web.bind.annotation.*;
import tech.riemann.ims.dto.request.ApplyInfo;
import tech.riemann.ims.dto.request.AuditApplyInfo;
import tech.riemann.ims.dto.request.ReviewDTO;
import tech.riemann.ims.dto.response.*;
import tech.riemann.ims.entity.material.*;
import tech.riemann.ims.enums.*;
import tech.riemann.ims.service.material.*;

import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

/**
 * @author mayong
 * @since 2024/11/28 15:29
 */
@RestController
@Tag(name = "Apply", description = "入库/出库 申请单管理")
@RequiredArgsConstructor
public class ApplyFormController {

    private final IApplyFormService applyFormService;
    private final IApplyDetailService applyDetailService;
    private final IMaterialStockDetailService materialStockDetailService;
    private final IMaterialService materialService;
    private final IStocktakingScanDetailService stocktakingScanDetailService;
    private final IStocktakingScanExceptionalDataService stocktakingScanExceptionalDataService;
    private final ITypeService typeService;
    private final IManualStockDetailService manualStockDetailService;

    @GetMapping("applies")
    @Operation(summary = "分页查询申请列表")
    public IPage<ApplyDTO> searchPage(
            @Parameter(description = "页码") @RequestParam(name = "page", required = false, defaultValue = "1") int page,
            @Parameter(description = "页面大小") @RequestParam(name = "size", required = false, defaultValue = "10") int size,
            @Parameter(description = "申请类型(1入库 3出库 4盘点)") @RequestParam(name = "applyType") Integer applyType,
            @Parameter(description = "物料类型") @RequestParam(name = "type", required = false) String type,
            @Parameter(description = "物料编码") @RequestParam(name = "code", required = false) String code,
            @Parameter(description = "物料名称") @RequestParam(name = "name", required = false) String name) {
        IPage<ApplyDTO> result = applyFormService.search(page, size, applyType, type, code, name);
        result.getRecords().forEach(item -> item.setTypeName(typeService.getTypeName(item.getType())));
        return result;
    }

    @GetMapping("audit-applies")
    @Operation(summary = "分页查询盘点列表")
    public IPage<ApplyForm> searchAuditPage(
            @Parameter(description = "页码") @RequestParam(name = "page", required = false, defaultValue = "1") int page,
            @Parameter(description = "页面大小") @RequestParam(name = "size", required = false, defaultValue = "10") int size,
            @Parameter(description = "盘点类型") @RequestParam(name = "auditType", required = false) AuditTypeEnum auditType,
            @Parameter(description = "盘点人") @RequestParam(name = "taker", required = false) String taker,
            @Parameter(description = "创建日期") @RequestParam(name = "createDate", required = false) LocalDateTime createDate,
            @Parameter(description = "审核状态") @RequestParam(name = "reviewResults", required = false) List<ReviewResultEnum> reviewResults) {
        return applyFormService.page(Page.of(page, size), Wrappers.<ApplyForm>lambdaQuery()
                .eq(auditType != null, ApplyForm::getAuditType, auditType)
                .eq(ApplyForm::getType, ApplyTypeEnum.AUDIT)
                .like(StringUtil.isNotBlank(taker), ApplyForm::getTaker, "%" + taker + "%")
                .ge(createDate != null, ApplyForm::getCreatedTime, createDate)
                .in(reviewResults != null && !reviewResults.isEmpty(), ApplyForm::getReviewResult, reviewResults)
                .orderByDesc(ApplyForm::getCreatedTime)
        );
    }

    @GetMapping("apply/{applyId}")
    @Operation(summary = "查询申请单详情")
    public ApplyInfo detail(@Parameter(description = "申请单ID") @PathVariable(name = "applyId") Integer applyId) {
        List<ApplyDetail> list = applyDetailService.list(Wrappers.<ApplyDetail>lambdaQuery().eq(ApplyDetail::getApplyId, applyId));
        return ApplyInfo.builder()
                .applyForm(applyFormService.getById(applyId))
                .applyDetails(list)
                .build();
    }

    @PutMapping("apply")
    @Operation(summary = "提交出入库申请单")
    @Transactional(rollbackFor = Exception.class)
    public void saveApply(@Validated @Parameter(description = "申请单") @RequestBody ApplyInfo applyInfo) {
        // 如果是采购入库 就新增数据，否则只改状态
        ApplyTypeEnum applyType = applyInfo.getApplyForm().getType();

        // 校验物料数量
        Map<Long, Integer> confirmMap = applyInfo.getApplyDetails().stream()
                .filter(ApplyDetail::getAssignRule)
                .collect(Collectors.toMap(ApplyDetail::getMaterialId, ApplyDetail::getConfirmQuantity));
        Map<Long, Long> dataMap = applyInfo.getDetailList().stream()
                .collect(Collectors.groupingBy(MaterialStockDetail::getMaterialId, Collectors.counting()));
        confirmMap.forEach((k, v) -> {
            Long l = dataMap.get(k);
            if (l == null || l.intValue() != v) {
                throw BizException.create("[" + k + "]物料数量不匹配");
            }
        });
        // 新增申请单
        applyFormService.save(applyInfo.getApplyForm());
        for (ApplyDetail applyDetail : applyInfo.getApplyDetails()) {
            applyDetail.setApplyId(applyInfo.getApplyForm().getId());
        }
        applyDetailService.saveBatch(applyInfo.getApplyDetails());

        if (applyType == ApplyTypeEnum.PURCHASE_RECEIPT) {
            // 新增物料
            if (!applyInfo.getDetailList().isEmpty()) {
                applyInfo.getDetailList().forEach(detail -> {
                    detail.setApplyId(applyInfo.getApplyForm().getId());
                    detail.setStatus(StockStatusEnum.IN);
                });
                materialStockDetailService.saveBatch(applyInfo.getDetailList());
            }
            //记录物料库存数据
            applyInfo.getApplyDetails().forEach(detail ->
                    materialService.update(Wrappers.<Material>lambdaUpdate()
                            .eq(Material::getId, detail.getMaterialId())
                            .setSql(detail.getAssignRule(),
                                    "m_stock = m_stock + " + detail.getConfirmQuantity())
                            .setSql(!detail.getAssignRule(),
                                    "m_stock = m_stock + " + detail.getQuantity())
                    ));
        } else if (applyType == ApplyTypeEnum.LOAN_OUT) {
            //修改物料状态
            if (!applyInfo.getDetailList().isEmpty()) {
                applyInfo.getDetailList().forEach(detail -> materialStockDetailService.update(Wrappers.<MaterialStockDetail>lambdaUpdate()
                        .set(MaterialStockDetail::getStatus, StockStatusEnum.OUT)
                        .eq(MaterialStockDetail::getBarcode, detail.getBarcode())));
            }
            //修改物料库存数据
            applyInfo.getApplyDetails().forEach(detail -> materialService.update(Wrappers.<Material>lambdaUpdate()
                    .eq(Material::getId, detail.getMaterialId())
                    .setSql(detail.getAssignRule(),
                            "m_stock = m_stock - " + detail.getConfirmQuantity())
                    .setSql(!detail.getAssignRule(),
                            "m_stock = m_stock - " + detail.getQuantity())
            ));
        } else if (applyType == ApplyTypeEnum.RETURN_RECEIPT) {
            //修改物料状态
            if (!applyInfo.getDetailList().isEmpty()) {
                applyInfo.getDetailList().forEach(detail -> materialStockDetailService.update(Wrappers.<MaterialStockDetail>lambdaUpdate()
                        .set(MaterialStockDetail::getStatus, StockStatusEnum.IN)
                        .eq(MaterialStockDetail::getBarcode, detail.getBarcode())));
            }
            //修改物料库存数据
            applyInfo.getApplyDetails().forEach(detail -> materialService.update(Wrappers.<Material>lambdaUpdate()
                    .eq(Material::getId, detail.getMaterialId())
                    .setSql(detail.getAssignRule(),
                            "m_stock = m_stock + " + detail.getConfirmQuantity())
                    .setSql(!detail.getAssignRule(),
                            "m_stock = m_stock + " + detail.getQuantity())
            ));
        } else if (applyType == ApplyTypeEnum.SCRAP_OUT) {
            //
        }
    }


    @PostMapping("audit-apply")
    @Operation(summary = "提交盘点申请单")
    @Transactional(rollbackFor = Exception.class)
    public void auditApply(@Validated @Parameter(description = "申请单") @RequestBody AuditApplyInfo applyInfo) {
        ApplyForm applyForm = applyInfo.to();
        applyForm.setType(ApplyTypeEnum.AUDIT);
        applyFormService.save(applyForm);
        List<Material> materials;
        if (applyInfo.getAuditType() == AuditTypeEnum.SCAN) {
            materials = materialService.listByIds(applyInfo.getIds());
            materials.forEach(material -> applyDetailService.save(ApplyDetail.builder()
                    .applyId(applyForm.getId())
                    .materialId(material.getId())
                    .quantity(material.getStock())
                    .build()));
        }

    }

    @PostMapping("manual/audit-apply")
    @Operation(summary = "提交人工盘点数据")
    public void submitManualStock(@Validated @Parameter(description = "人工盘点数据") @RequestBody List<ManualStockDetail> details) {
        if (details.isEmpty()) {
            return;
        }
        manualStockDetailService.saveBatch(details);
        //修改状态
        applyFormService.update(Wrappers.<ApplyForm>lambdaUpdate()
                .eq(ApplyForm::getId, details.get(0).getApplyId())
                .set(ApplyForm::getReviewResult, ReviewResultEnum.WAIT_REVIEW));
    }

    @GetMapping("manual/data/{applyId}")
    @Operation(summary = "获取人工盘点数据")
    public List<ManualStockDTO> getManualStock(@Parameter(description = "申请单ID") @PathVariable(name = "applyId") Long applyId) {
        List<ManualStockDTO> result = manualStockDetailService.getByApplyId(applyId);
        result.forEach(item -> item.setTypeName(typeService.getTypeName(item.getType())));
        return result;
    }


    @GetMapping("wait-scan-data/{applyId}")
    @Operation(summary = "获取盘点的待扫码信息")
    public List<WaitScanInfo> getWaitScanData(@Parameter(description = "申请单ID") @PathVariable(name = "applyId") Long applyId) {
        return applyDetailService.getWaitScanData(applyId);
    }

    @PostMapping("scan-data/{applyId}")
    @Operation(summary = "提交盘点扫码数据")
    @Transactional(rollbackFor = Exception.class)
    public void saveScanData(@Parameter(description = "扫码数据列表") @RequestBody List<StocktakingScanDetail> scanData,
                             @Parameter(description = "申请单ID") @PathVariable(name = "applyId") Long applyId) {
        //保存扫码数据
        stocktakingScanDetailService.saveBatch(scanData);
        //修改状态
        applyFormService.update(Wrappers.<ApplyForm>lambdaUpdate()
                .set(ApplyForm::getReviewResult, ReviewResultEnum.WAIT_SUBMIT)
                .eq(ApplyForm::getId, applyId));
        new Thread(() -> comparisonData(scanData, applyId)).start();
    }

    @GetMapping("comparison-result/{applyId}")
    @Operation(summary = "获取盘点结果数据")
    public ComparisonResDTO getComparisonRes(@Parameter(description = "申请单ID") @PathVariable(name = "applyId") String applyId) {
        List<ApplyDetail> applyDetails = applyDetailService.list(Wrappers.<ApplyDetail>lambdaQuery().eq(ApplyDetail::getApplyId, applyId));
        List<Long> materialIds = applyDetails.stream().map(ApplyDetail::getMaterialId).toList();

        List<MaterialStockDetail> stockDetails = materialStockDetailService.list(Wrappers.<MaterialStockDetail>lambdaQuery().in(MaterialStockDetail::getMaterialId, materialIds));
        Map<Long, Long> stockMap = stockDetails.stream().collect(Collectors.groupingBy(MaterialStockDetail::getMaterialId, Collectors.counting()));

        List<StocktakingScanDetail> scanDetails = stocktakingScanDetailService.list(Wrappers.<StocktakingScanDetail>lambdaQuery().eq(StocktakingScanDetail::getApplyId, applyId));
        Map<Long, Long> scanMap = scanDetails.stream().collect(Collectors.groupingBy(StocktakingScanDetail::getMaterialId, Collectors.counting()));

        List<StockDetailInfo> infoList = new ArrayList<>();
        materialIds.forEach(id -> {
            Material material = materialService.getById(id);
            infoList.add(StockDetailInfo.builder()
                    .materialId(material.getId())
                    .materialName(material.getName())
                    .materialCode(material.getCode())
                    .stockTotal(material.getStock())
                    .stockTotalScan(stockMap.getOrDefault(id, 0L))
                    .stockReal(scanMap.getOrDefault(id, 0L))
                    .build());
        });

        List<StocktakingScanExceptionalData> dataList = stocktakingScanExceptionalDataService.selectData(applyId);

        return ComparisonResDTO.builder()
                .details(infoList)
                .exceptionalData(dataList)
                .build();
    }

    @PutMapping("handle-exceptional-data/{applyId}")
    @Operation(summary = "处理盘点异常数据")
    public void updateReviewResult(@Parameter(description = "申请单ID") @PathVariable(name = "applyId") String applyId,
                                   @Parameter(description = "异常数据列表") @RequestBody List<StocktakingScanExceptionalData> dataList) {
        dataList.forEach(data -> stocktakingScanExceptionalDataService.update(Wrappers.<StocktakingScanExceptionalData>lambdaUpdate()
                .eq(StocktakingScanExceptionalData::getId, data.getId())
                .set(StocktakingScanExceptionalData::getExceptionHandle, data.getExceptionHandle())
                .set(StocktakingScanExceptionalData::getRemark, data.getRemark())));
        applyFormService.update(Wrappers.<ApplyForm>lambdaUpdate()
                .set(ApplyForm::getReviewResult, ReviewResultEnum.WAIT_REVIEW)
                .eq(ApplyForm::getId, applyId));

    }

    @PostMapping("submit-review")
    @Operation(summary = "提交审核")
    @Transactional(rollbackFor = Exception.class)
    public void submitReview(@Parameter(description = "异常数据列表") @RequestBody ReviewDTO reviewDTO) {
        applyFormService.update(Wrappers.<ApplyForm>lambdaUpdate()
                .set(ApplyForm::getReviewResult, reviewDTO.getReviewResult())
                .set(reviewDTO.getReviewResult() == ReviewResultEnum.PASS, ApplyForm::getConfirm, Boolean.TRUE)
                .set(ApplyForm::getReviewRemark, reviewDTO.getRemark())
                .eq(ApplyForm::getId, reviewDTO.getApplyId()));
        ApplyForm byId = applyFormService.getById(reviewDTO.getApplyId());
        if (byId.getAuditType() == AuditTypeEnum.MANUAL) {
            if (reviewDTO.getReviewResult() != ReviewResultEnum.PASS) {
                return;
            }
            List<ManualStockDetail> details = manualStockDetailService.list(Wrappers.<ManualStockDetail>lambdaQuery()
                    .eq(ManualStockDetail::getApplyId, reviewDTO.getApplyId()));
            details.forEach(detail -> materialService.update(Wrappers.<Material>lambdaUpdate()
                    .set(Material::getStock, detail.getManualStock())
                    .eq(Material::getId, detail.getMaterialId())));
            return;
        }
        if (reviewDTO.getReviewResult() == ReviewResultEnum.PASS) {
            //修改库存明细状态和库存数量
            List<StocktakingScanExceptionalData> list = stocktakingScanExceptionalDataService.list(Wrappers.<StocktakingScanExceptionalData>lambdaQuery()
                    .eq(StocktakingScanExceptionalData::getApplyId, reviewDTO.getApplyId()));
            list.forEach(data -> {
                // 新入库 归还 丢失 需要处理
                if (data.getExceptionHandle() == HandleEnum.MARK_NEW) {
                    // 新增库存明细 和 增加库存数量
                    materialStockDetailService.save(MaterialStockDetail.builder()
                            .applyId(null)
                            .materialId(data.getMaterialId())
                            .barcode(data.getBarcode())
                            .status(StockStatusEnum.IN)
                            .build());
                    materialService.update(Wrappers.<Material>lambdaUpdate()
                            .setSql("m_stock = m_stock + " + 1)
                            .eq(Material::getId, data.getMaterialId()));
                } else if (data.getExceptionHandle() == HandleEnum.MARK_RETURN || data.getExceptionHandle() == HandleEnum.MARK_LOST) {
                    materialStockDetailService.update(Wrappers.<MaterialStockDetail>lambdaUpdate()
                            .set(MaterialStockDetail::getStatus, data.getExceptionHandle() == HandleEnum.MARK_RETURN
                                    ? StockStatusEnum.IN : StockStatusEnum.LOST)
                            .eq(MaterialStockDetail::getBarcode, data.getBarcode()));
                    materialService.update(Wrappers.<Material>lambdaUpdate()
                            .setSql(data.getExceptionHandle() == HandleEnum.MARK_RETURN, "m_stock = m_stock + " + 1)
                            .setSql(data.getExceptionHandle() == HandleEnum.MARK_LOST, "m_stock = m_stock - " + 1)
                            .eq(Material::getId, data.getMaterialId()));
                }
            });
        }
    }

    // 比较数据 生成差异数据
    public void comparisonData(List<StocktakingScanDetail> scanData, Long applyId) {
        //生成map比较
        Map<Long, List<String>> scanMap = scanData.stream().collect(Collectors.groupingBy(StocktakingScanDetail::getMaterialId,
                Collectors.mapping(StocktakingScanDetail::getBarcode, Collectors.toList())));
        Map<Long, List<MaterialStockDetail>> stockMap = getStockMap(applyId);
        // 比较两个map,以库存为准,生成差异数据
        List<StocktakingScanExceptionalData> exceptionalDataList = new ArrayList<>();
        scanMap.forEach((k, v) -> {
            List<MaterialStockDetail> stockList = stockMap.get(k);
            List<String> scanList = scanMap.get(k);
            // 比较两个list,找出库存存在但扫码不存在的条形码 或者 库存不存在但扫码存在的条形码
            stockList.forEach(detail -> {
                if (scanList.contains(detail.getBarcode()) && detail.getStatus() != StockStatusEnum.IN) {
                    // 扫码和库存都存在，但是库存的状态不对
                    exceptionalDataList.add(StocktakingScanExceptionalData.builder()
                            .barcode(detail.getBarcode())
                            .applyId(applyId)
                            .materialId(detail.getMaterialId())
                            .exceptionReason(StocktakingExceptonalStatusEnum.SOCK_OUT_BUT_SCAN_EXIST)
                            .build());
                }
                if (!scanList.contains(detail.getBarcode()) && detail.getStatus() == StockStatusEnum.IN) {
                    //库房存在，但扫码不存在
                    exceptionalDataList.add(StocktakingScanExceptionalData.builder()
                            .barcode(detail.getBarcode())
                            .materialId(detail.getMaterialId())
                            .applyId(applyId)
                            .exceptionReason(StocktakingExceptonalStatusEnum.SOCK_IN_BUT_SCAN_NOT_EXIST)
                            .build());
                }
            });
            // 找出扫码存在但库存不存在的条形码
            scanList.forEach(barcode -> {
                if (stockList.stream().noneMatch(detail -> detail.getBarcode().equals(barcode))) {
                    // 扫码存在，但库存不存在
                    exceptionalDataList.add(StocktakingScanExceptionalData.builder()
                            .barcode(barcode)
                            .materialId(k)
                            .applyId(applyId)
                            .exceptionReason(StocktakingExceptonalStatusEnum.SOCK_NOT_EXIST_BUT_SCAN_EXIST)
                            .build());
                }
            });
        });
        // 保存差异数据
        stocktakingScanExceptionalDataService.saveBatch(exceptionalDataList);
    }

    // 盘点物料库存实际数据
    private Map<Long, List<MaterialStockDetail>> getStockMap(Long applyId) {
        ApplyForm applyForm = applyFormService.getById(applyId);
        List<Long> materialIds = new ArrayList<>();
        if (applyForm.getAuditType() == AuditTypeEnum.MANUAL) {
            // 部分盘点
            List<ApplyDetail> applyDetails = applyDetailService.list(Wrappers.<ApplyDetail>lambdaQuery().eq(ApplyDetail::getApplyId, applyId));
            materialIds = applyDetails.stream().map(ApplyDetail::getMaterialId).toList();
        }
        //查询所有盘点物料数量
        List<MaterialStockDetail> stockDetails = materialStockDetailService.list(Wrappers.<MaterialStockDetail>lambdaQuery()
                .in(!materialIds.isEmpty(), MaterialStockDetail::getMaterialId, materialIds));

        // 返回map结构 key 为物料id，value 该物料的所有条形码结合
        return stockDetails.stream().collect(Collectors.groupingBy(MaterialStockDetail::getMaterialId));
    }
}
