package tech.riemann.ims.controller.platform.material;

import club.zhcs.lina.starter.exception.BizException;
import com.baomidou.mybatisplus.core.metadata.IPage;
import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.google.common.cache.CacheBuilder;
import com.google.common.cache.CacheLoader;
import com.google.common.cache.LoadingCache;
import io.swagger.v3.oas.annotations.Operation;
import io.swagger.v3.oas.annotations.Parameter;
import io.swagger.v3.oas.annotations.tags.Tag;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.nutz.lang.Strings;
import org.nutz.lang.random.R;
import org.springframework.http.HttpStatus;
import org.springframework.validation.annotation.Validated;
import org.springframework.web.bind.annotation.*;
import tech.riemann.ims.entity.material.Material;
import tech.riemann.ims.service.material.IMaterialService;
import tech.riemann.ims.service.material.ITypeService;
import tech.riemann.ims.utils.BarcodeUtil;

import javax.annotation.Nonnull;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;

/**
 * @author mayong
 * @since 2024/11/27 15:11
 */
@RestController
@Slf4j
@Tag(name = "Material", description = "物料管理")
@RequiredArgsConstructor
public class MaterialController {

    private final IMaterialService materialService;
    private final ITypeService typeService;

    LoadingCache<String, String> cache = CacheBuilder.newBuilder()
            .concurrencyLevel(8)
            .expireAfterAccess(10, TimeUnit.MINUTES)
            .initialCapacity(10)
            .maximumSize(300)
            .recordStats()
            .build(new CacheLoader<String, String>() {

                @Override
                @Nonnull
                public String load(@Nonnull String key) throws Exception {
                    Material material = materialService.getOne(Wrappers.<Material>lambdaQuery().eq(Material::getCode, key));
                    return material == null ? "" : material.getCode();
                }
            });

    @GetMapping("materials")
    @Operation(summary = "分页查询物料列表")
    public IPage<Material> materials(
            @Parameter(description = "页码") @RequestParam(required = false, defaultValue = "1") int page,
            @Parameter(description = "页面大小") @RequestParam(required = false, defaultValue = "10") int size,
            @Parameter(description = "类型") @RequestParam(required = false, defaultValue = "") Integer type,
            @Parameter(description = "搜索关键词") @RequestParam(required = false, defaultValue = "") String key) {
        boolean hasLike = Strings.isNotBlank(key);
        key = String.format("%%%s%%", key);
        Page<Material> result = materialService.page(Page.of(page, size), Wrappers.<Material>lambdaQuery()
                .like(hasLike, Material::getName, key)
                .or()
                .like(hasLike, Material::getCode, key)
                .or()
                .like(hasLike, Material::getType, key)
                .orderByDesc(Material::getUpdatedTime));
        result.getRecords().forEach(item -> item.setTypeName(typeService.getTypeName(item.getType())));
        return result;
    }

    @PostMapping("material/list")
    @Operation(summary = "查询所有物料列表")
    public List<Material> all(@Parameter(description = "类型") @RequestParam(required = false, defaultValue = "") String type,
                              @Parameter(description = "是否赋码") @RequestParam(required = false) Boolean assignRule) {
        String[] types = StringUtils.isNotBlank(type) ? type.split(",") : new String[0];
        List<Material> materials = materialService.queryLikeRight(List.of(types));
        if (assignRule != null) {
            materials.removeIf(item -> assignRule != item.getAssignRule());
        }
        materials.forEach(item -> item.setTypeName(typeService.getTypeName(item.getType())));
        return materials;
    }

    @GetMapping("material/{id}")
    @Operation(summary = "物料详情")
    public Material detail(@Parameter(description = "物料id", required = true) @PathVariable long id) {
        return materialService.getById(id);
    }

    @PutMapping("material")
    @Operation(summary = "增加/编辑物料")
    public Material saveOrUpdateMaterial(@Validated @Parameter(description = "物料信息") @RequestBody Material material) {
        if (material.getId() == null || material.getId() <= 0) {
            material.setCode(getCode());
            material.setStock(0);
            if (materialService.save(material)) {
                return material;
            } else {
                throw BizException.create("保存物料失败!");
            }
        } else {
            if (materialService.update(material, Wrappers.<Material>lambdaUpdate().eq(Material::getId, material.getId()))) {
                return material;
            } else {
                throw BizException.create("更新物料失败!");
            }
        }
    }

    private String getCode() {
        String code;
        while (true) {
            code = String.valueOf(R.random(1000, 9999));
            try {
                // 如果代码已经存在，继续循环
                if (Strings.isBlank(cache.get(code))) {
                    cache.put(code, code);
                    return code;  // 找到合适的代码后返回
                }
            } catch (ExecutionException e) {
                log.error("获取物料编码失败", e);
            }
        }
    }

    @DeleteMapping("material/{id}")
    @Operation(summary = "删除物料")
    @ResponseStatus(HttpStatus.OK)
    public void deleteMaterial(@Parameter(description = "物料id", required = true) @PathVariable long id) {
        if (!materialService.removeById(id)) {
            throw BizException.create("删除物料失败!");
        }
    }

    // 批量生成条形码
    @PostMapping("material/{id}/generate-barcodes")
    @Operation(summary = "批量生成条形码")
    public List<String> generateBarcodes(@Parameter(description = "物料id", required = true) @PathVariable Long id,
                                         @Parameter(description = "条形码数量", required = true) @RequestParam(name = "count") int count) {
        List<String> res = new ArrayList<>();
        Material byId = materialService.getById(id);
        if (byId != null && count > 0) {
            String code = byId.getCode();
            while (count-- > 0) {
                res.add(code + BarcodeUtil.generateBarcode());
            }
        }
        log.info("生成条形码成功，条形码列表：{}", res);
        return res;
    }
}
